﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionaryTracking {
  [Item("Genealogy graph", "Generic class for tracking the evolution of objects.")]
  public class GenealogyGraph<T> : NamedItem, IGenealogyGraph<T> where T : class {
    private readonly Dictionary<T, GenealogyGraphNode> _dictionary;

    public GenealogyGraph() {
      _dictionary = new Dictionary<T, GenealogyGraphNode>();
    }

    public GenealogyGraph(GenealogyGraph<T> g) {
      _dictionary = new Dictionary<T, GenealogyGraphNode>(g._dictionary);
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GenealogyGraph<T>(this, cloner);
    }

    public override Image ItemImage {
      get { return Common.Resources.VSImageLibrary.Graph; }
    }

    [StorableConstructor]
    private GenealogyGraph(bool serializing)
      : base(serializing) {
    }

    private GenealogyGraph(GenealogyGraph<T> original, Cloner cloner)
      : base(original, cloner) {
      _dictionary = new Dictionary<T, GenealogyGraphNode>(original._dictionary);
    }

    public bool HasNode(T t) {
      return _dictionary.ContainsKey(t);
    }

    public GenealogyGraphNode GetNode(T t) {
      return HasNode(t) ? _dictionary[t] : null;
    }

    public IEnumerable<T> Keys {
      get { return _dictionary.Keys; }
    }

    public IEnumerable<GenealogyGraphNode> Values {
      get { return _dictionary.Values; }
    }

    public bool IsEmpty {
      get { return !_dictionary.Any(); }
    }

    public bool Any(Func<KeyValuePair<T, GenealogyGraphNode>, bool> predicate) {
      return _dictionary.Any(predicate);
    }

    public void Clear() {
      _dictionary.Clear();
    }

    /// <summary>
    /// Adds a node representing an individual
    /// </summary>
    /// <param name="t">The symbolic expression tree</param>
    /// <param name="quality">The quality value </param>
    /// <param name="label">The node label</param>
    /// <param name="rank">The node rank</param>
    /// <param name="elite">Specifies if this node is an elite</param>
    public void AddNode(T t, double quality = 0.0, string label = "", double rank = 0, bool elite = false, int cutIndex = -1) {
      if (HasNode(t)) return;
      _dictionary[t] = new GenealogyGraphNode { Data = t, Quality = quality, Label = label, IsElite = elite, Rank = rank, CutpointIndex = cutIndex };
    }

    /// <summary>
    ///  more of a low level method to minimize memory usage when creating and returning subgraphs
    /// </summary>
    /// <param name="n">The node to be added</param>
    public void AddNode(GenealogyGraphNode n) {
      var t = (T)n.Data;
      if (HasNode(t)) return;
      _dictionary[t] = n;
    }

    /// <summary>
    /// Remove a node from the graph along with edges associated with it
    /// </summary>
    /// <param name="t">The graph node</param>
    public void RemoveNode(T t) {
      if (!_dictionary.ContainsKey(t)) return;
      GenealogyGraphNode n = _dictionary[t];
      if (n.InEdges != null) {
        foreach (var e in n.InEdges.Where(e => e.Target.OutEdges != null)) {
          e.Target.OutEdges.RemoveAll(arc => arc.Target == n);
          if (!e.Target.OutEdges.Any())
            e.Target.OutEdges = null; // set to null to be a little more memory efficient
        }
        n.InEdges = null;
      }
      if (n.OutEdges != null) {
        foreach (var e in n.OutEdges.Where(e => e.Target.InEdges != null)) {
          e.Target.InEdges.RemoveAll(arc => arc.Target == n);
          if (!e.Target.InEdges.Any())
            e.Target.InEdges = null; // set to null to be a little more memory efficient
        }
        n.OutEdges = null;
      }
      _dictionary.Remove(t);
    }

    public double AverageDegree {
      get { return Values.Average(x => x.Degree); }
    }

    /// <summary>
    /// Adds a graph arc from a to b
    /// </summary>
    /// <param name="a"></param>
    /// <param name="b"></param>
    /// <param name="label"></param>
    public void AddArc(T a, T b, int opType = 0) {
      GenealogyGraphNode src, dest;
      if (!HasNode(a)) {
        src = new GenealogyGraphNode { Data = a };
        _dictionary[a] = src;
      } else {
        src = _dictionary[a];
      }
      if (!HasNode(b)) {
        dest = new GenealogyGraphNode { Data = b };
        _dictionary[b] = dest;
      } else {
        dest = _dictionary[b];
      }
      // add forward and reverse arcs
      src.AddForwardArc(dest, opType);
      dest.AddReverseArc(src, opType);
    }

    public void AddArcs(T[] a, T b, int opType = 0) {
      GenealogyGraphNode src, dest;
      if (!HasNode(b)) {
        dest = new GenealogyGraphNode { Data = b };
        _dictionary[b] = dest;
      } else {
        dest = _dictionary[b];
      }
      foreach (var o in a) {
        if (!HasNode(o)) {
          src = new GenealogyGraphNode { Data = o };
          _dictionary[o] = src;
        } else {
          src = _dictionary[o];
        }
        src.AddForwardArc(dest, opType);
        dest.AddReverseArc(src, opType);
      }
    }
  }

  public class GenealogyGraphNode : IComparable {
    public string Id { get; private set; }
    public string Label { get; set; }
    public bool IsElite { get; set; }
    public object Data { get; set; }
    public double Quality { get; set; }
    public double Rank { get; set; }
    public int CutpointIndex { get; set; }
    public List<GenealogyGraphArc> InEdges { get; set; }
    public List<GenealogyGraphArc> OutEdges { get; set; }

    public int Degree { get { return (InEdges == null ? 0 : InEdges.Count) + (OutEdges == null ? 0 : OutEdges.Count); } }

    public GenealogyGraphNode() { Id = Guid.NewGuid().ToString(); }

    public GenealogyGraphNode(GenealogyGraphNode node) {
      Id = Guid.NewGuid().ToString();
      Rank = node.Rank;
      Quality = node.Quality;
      IsElite = node.IsElite;
      Label = node.Label;
      Data = node.Data;
      InEdges = node.InEdges;
      OutEdges = node.OutEdges;
      CutpointIndex = node.CutpointIndex;
    }

    /// <summary>
    /// Returns all the ancestors of the current node
    /// </summary>
    /// <returns></returns>
    public IEnumerable<GenealogyGraphNode> Ancestors() {
      var nodes = new HashSet<GenealogyGraphNode> { this };
      int offset = 0;
      int c0 = 1;
      while (offset != c0) {
        int c1 = c0;
        for (int i = offset; i != c1; ++i) {
          if (nodes.ElementAt(i).InEdges == null) continue;
          foreach (var p in nodes.ElementAt(i).InEdges.Select(e => e.Target)) {
            nodes.Add(p);
            yield return p;
          }
        }
        offset = c1;
        c0 = nodes.Count;
      }
    }

    /// <summary>
    /// Returns all the descendants of the current node (identical to above method except for using the OutEdges)
    /// </summary>
    /// <returns></returns>
    public IEnumerable<GenealogyGraphNode> Descendants() {
      var nodes = new HashSet<GenealogyGraphNode> { this };
      int offset = 0;
      int c0 = 1;
      while (offset != c0) {
        int c1 = c0;
        for (int i = offset; i != c1; ++i) {
          if (nodes.ElementAt(i).OutEdges == null) continue;
          foreach (var p in nodes.ElementAt(i).OutEdges.Select(e => e.Target)) {
            nodes.Add(p);
            yield return p;
          }
        }
        offset = c1;
        c0 = nodes.Count;
      }
    }

    public void AddForwardArc(GenealogyGraphNode target, int opType) {
      var e = new GenealogyGraphArc { Target = target, OperatorType = opType };
      if (OutEdges == null) OutEdges = new List<GenealogyGraphArc> { e };
      else OutEdges.Add(e);
    }

    public void AddReverseArc(GenealogyGraphNode target, int opType) {
      var e = new GenealogyGraphArc { Target = target, OperatorType = opType };
      if (InEdges == null) InEdges = new List<GenealogyGraphArc> { e };
      else InEdges.Add(e);
    }

    // comparable
    // may seem weird, it is actually implemented so higher quality means "less than" in terms of ordering
    // if quality is equal, the elite node takes precedence
    public int CompareTo(object obj) {
      if (obj == null) return 1;
      var other = obj as GenealogyGraphNode;
      if (other == null) return 1;
      if (Quality.Equals(other.Quality)) {
        if (IsElite) return -1;
        return other.IsElite ? 1 : -1;
      }
      return other.Quality.CompareTo(Quality);
    }
  }

  /// <summary>
  /// An arc is an edge along with an orientation
  /// </summary>
  public class GenealogyGraphArc {
    public GenealogyGraphNode Target { get; set; }
    // OperationType is an identifier for the genetic operation (mutation, crossover) that a graph arc will represent 
    // So basically it describes which operator was applied to an individual/graph node (the one emitting the arc), 
    // while Target represents the end result of that operator (node ==GeneticOperation==> Target)
    public int OperatorType { get; set; }
    // these two fields are not used, but they might be useful later 
    public string Label { get; set; } // might be useful later
    public double Weight { get; set; } // might be useful later
  }
}
